//*****************************************************************************
//* lcdif.h | lcdif.c
//*
//* Info: 			Treiberschicht zur Ansteuerung des LH155-Displaycontrollers
//*
//* Author(s): 		Fabian Cordes
//* Copyright: 		(c) 2009 Projekt ELE21 
//*
//*****************************************************************************


// TODO: Define used clock rate
#define F_CPU 16000000UL
#include <util/delay.h>

#include "..\ports.h"
#include "lcdif.h"
#include "lcdsettings.h"


// Internal prototypes for low level functions
inline void lcdActivateWRB (void);
inline void lcdDeactivateWRB (void);
inline void lcdActivateRDB (void);
inline void lcdDeactivateRDB (void);
inline void lcdActivateCSB (void);
inline void lcdDeactivateCSB (void);
inline void lcdActivateReset (void);
inline void lcdDeactivateReset (void);
inline void lcdActivateRS (void);
inline void lcdDeactivateRS (void);

// Internal prototypes for driver level functions
void setYAddress ( uint8_t dataByte);
void setXAddress ( uint8_t dataByte);
void setStartAddress (uint8_t dataByte);
void setNLineAlternate (uint8_t dataByte );
void setRegister (uint8_t dataByte);



// various hardcoded delays with high precision (1/2/4 clk)
static inline void delay_1clk ( void )
{
	asm volatile 	(	"nop\n\t"
					);
}

static inline void delay_2clk ( void )
{
	asm volatile 	(	"nop\n\t"
						"nop\n\t"
					);
}

static inline void delay_4clk ( void )
{
	asm volatile 	(	"nop\n\t"
						"nop\n\t"
						"nop\n\t"
						"nop\n\t"
					);
}

// very long delay for debugging
static inline void delay_long ( void )
{
	// delay 1 msec
	_delay_ms(1);
}

/*******************************
Display Initialization
*******************************/

void initLCD ( void )
{
	// Drive display active and wait a moderate amount of time
	lcdDeactivateReset();
	_delay_ms(100);	

	// Begin initialization routine
	writeLCDCtrl( 0, LCD_REFLAG1 ); 			// set RE flag off
	writeLCDCtrl( 0, LCD_SET_START_ADDRESS );	// starting line = 0
	writeLCDCtrl( 1, LCD_SET_NLINEALT );		// n-line alternate
	writeLCDCtrl( (1<< LCD_DCTRL1_ALLON), LCD_DISPCTRL1); 
												// All pixels on
	writeLCDCtrl( 0, LCD_DISPCTRL2);			// no changes for DISPCTRL2
	writeLCDCtrl( ((1<< LCD_ICTRL1_INCX)|(1 << LCD_ICTRL1_INCY)),LCD_INCCTRL1);
												// Auto increment X and Y
	writeLCDCtrl( (1<< LCD_PCTRL1_PWRON), LCD_PWRCTRL1);
												// Power on
	writeLCDCtrl( 	(1<< LCD_PCTRL2_DVOL0) |
				 	(1<< LCD_PCTRL2_DVOL1) |
					(1<< LCD_PCTRL2_DVOL2) |
					(1<< LCD_PCTRL2_DVOL3),
					LCD_PWRCTRL2 );				// Electronic Volume off
	writeLCDCtrl( 0, LCD_PWRCTRL3 );			// RE = 0 -> set graphics mode
	writeLCDCtrl( 1, LCD_REFLAG1 );				// Set RE flag on
	writeLCDCtrl( 0, LCD_PWRCTRL3 );			// Duty cycle 1/64
	
	writeLCDCtrl( 0, LCD_SET_X_ADDRESS );		// Reset X counter
	writeLCDCtrl( 0, LCD_SET_Y_ADDRESS );		// Reset Y counter
	

	// Wait a little while
	_delay_ms(100);		
	
	// Display can be used now
	return;
}


/*******************************
Writing data packet to LCD
*******************************/

// Generic write to lcd. Does no attempt to specify where data is written, 
// this needs to be done prior to calling using writeLCDCtrl(). 
// Function is, however, very fast. 

void writeLCDData ( uint8_t dataByte )
{
	lcdDeactivateReset();
	// Activate CSB
	lcdActivateCSB();

	// Respect signal timings
	delay_tAS8();

	// Activate WRB, enable outputs
	lcdActivateWRB();
	// Put data on port, which must be output by now
	LCDDATA = dataByte;

	// Respect signal timings
	delay_tWRW();
	
	// Deactivate WRB
	lcdDeactivateWRB();

	// Respect signal timings
	delay_tAH8();
	
	// Deactivate CSB
	lcdDeactivateCSB();
	
	return;
}


/*******************************
Writing to control registers
*******************************/

// Configureable write to lcd. Used to interface extra functions in the
// controller IC. Do not use for generic writes!
// As ctrlByte parameter, use the predefined opcodes (see lcdif.h)

void writeLCDCtrl ( uint8_t dataByte, uint8_t ctrlByte )
{
	// Examine control parameter
	switch (ctrlByte)
	{
		// Special-case handling since two write accesses are needed
		case LCD_SET_Y_ADDRESS:
			setYAddress( dataByte );
		break;
				
		// Special-case handling since two write accesses are needed
		case LCD_SET_X_ADDRESS:
			setXAddress( dataByte );
		break;

		// Special-case handling since two write accesses are needed
		case LCD_SET_START_ADDRESS:
			setStartAddress ( dataByte );
		break;

		// Special-case handling since two write accesses are needed
		case LCD_SET_NLINEALT:
			setNLineAlternate ( dataByte );
		break;

		default:
			// Generic control register is targetet by ctrlByte		
			// Assemble dataByte
			// Use lowest 4 bit of dataByte and shift ctrlByte into the higher 4 bits
			dataByte &= ((1<<D4) | (1<<D5) | (1<<D6) | (1<<D7));
			dataByte |= (ctrlByte & ((1<<D0) | (1<<D1) | (1<<D2) | (1<<D3))) << 4;
			// Set up control register
			setRegister (dataByte);

		break;
	}
	return;
}


/*******************************
Handling for extended functions
*******************************/

// Sets Y address for future reads/writes
// See LH155 Datasheet for documentation on the protocol
void setYAddress ( uint8_t dataByte )
{
	uint8_t dataByteLo = 0;
	uint8_t dataByteHi = 0;
	// split Address into two 4-bit pieces
	// ...the lower 4 bits
	dataByteLo = dataByte & ((1<<D0) | (1<<D1) | (1<<D2) | (1<<D3));
	// Command for setting Y(lo) is 0x2h
	dataByteLo = dataByteLo | (1<<D5);
	
	// ...and the Upper 4 bits
	dataByteHi = dataByte & ((1<<D4) | (1<<D5) | (1<<D6) | (1<<D7));
	// shift right by 4 bits
	dataByteHi = dataByteHi >> 4;
	// Command for setting Y(high) is 0x3h
	dataByteHi = dataByteHi | ((1<<D4) | (1<<D5));

	// Set up control lines
	// set RS line
	lcdActivateRS();

	// perform two writes, one for each part of dataByte
	writeLCDData(dataByteLo);
	writeLCDData(dataByteHi);

	// unset RS line
	lcdDeactivateRS();

	// done here
	return;
}

// Sets X address for future reads/writes
// Same as setting Y address, exept for different nibbles.
//
// Warning: Datasheet is missing the command for setting the upper 
// part of the address since it describes a similar, but different IC
// See LH155 Datasheet for documentation on the protocol
void setXAddress ( uint8_t dataByte )
{
	uint8_t dataByteLo = 0;
	uint8_t dataByteHi = 0;
	// split Address into two 4-bit pieces
	// ...the lower 4 bits
	dataByteLo = dataByte & ((1<<D0) | (1<<D1) | (1<<D2) | (1<<D3));
	// Command for setting X(low) is 0x0h
	dataByteLo = 0;
	
	// ...and the upper 4 bits
	dataByteHi = dataByte & ((1<<D4) | (1<<D5) | (1<<D6) | (1<<D7));
	// shift right by 4 bits
	dataByteHi = dataByteHi >> 4;
	// Command for setting X(high) is 0x1h
	dataByteHi = dataByteHi | (1<<D1);

	// Set up control lines
	// set RS line
	lcdActivateRS();

	// perform two writes, one for each part of dataByte
	writeLCDData(dataByteLo);
	writeLCDData(dataByteHi);

	// unset RS line
	lcdDeactivateRS();

	// done here
	return;
}

// Sets starting address for the first displayed line
// See LH155 Datasheet for documentation on the protocol
void setStartAddress ( uint8_t dataByte )
{
	uint8_t dataByteLo = 0;
	uint8_t dataByteHi = 0;
	// split Address into two 4-bit pieces
	// ...the lower 4 bits
	dataByteLo = dataByte & ((1<<D0) | (1<<D1) | (1<<D2) | (1<<D3));
	// Command for setting start address(low) is 0x4h
	dataByteLo = (1<<D6);
	
	// ...and the upper 4 bits
	dataByteHi = dataByte & ((1<<D4) | (1<<D5) | (1<<D6) | (1<<D7));
	// shift right by 4 bits
	dataByteHi = dataByteHi >> 4;
	// Command for setting Start Address(high) is 0x5h
	dataByteHi = dataByteHi | ((1<<D4) | (1<<D6));

	// Set up control lines
	// set RS line
	lcdActivateRS();

	// perform two writes, one for each part of dataByte
	writeLCDData(dataByteLo);
	writeLCDData(dataByteHi);

	// unset RS line
	lcdDeactivateRS();

	// done here
	return;
}

void setNLineAlternate (uint8_t dataByte )
{
	uint8_t dataByteLo = 0;
	uint8_t dataByteHi = 0;
	// split Address into two 4-bit pieces
	// ...the lower 4 bits
	dataByteLo = dataByte & ((1<<D0) | (1<<D1) | (1<<D2) | (1<<D3));
	// Command for setting alternate(low) is 0x6h
	dataByteLo = ((1<<D5) | (1<<D6));
	
	// ...and the upper 4 bits
	dataByteHi = dataByte & ((1<<D4) | (1<<D5) | (1<<D6) | (1<<D7));
	// shift right by 4 bits
	dataByteHi = dataByteHi >> 4;
	// Command for setting alternate(high) is 0x7h
	dataByteHi = dataByteHi | ((1<<D4) | (1<<D5) | (1<<D6));

	// Set up control lines
	// set RS line
	lcdActivateRS();

	// perform two writes, one for each part of dataByte
	writeLCDData(dataByteLo);
	writeLCDData(dataByteHi);

	// unset RS line
	lcdDeactivateRS();
}

// Sets starting address for the first displayed line
// See LH155 Datasheet for documentation on the protocol
void setRegister ( uint8_t dataByte )
{
	// set RS line
	lcdActivateRS();

	// dataByte-4..7 address the page, dataByte-0..3 set the flags
	// perform write
	writeLCDData(dataByte);

	// unset RS line
	lcdDeactivateRS();

	// done here
	return;
}





/*******************************
Low level wrapper functions for IO
*******************************/

// needed to activate and deativate the lcd write data bit ////////////////////
inline void lcdActivateWRB (void)
{
	// always switch on outputs before activating the write data bit
	// pull up RDB (LO act.) 
	LCDCTRL |= (1 << RDB); // asm volatile ("sbr LCDCTRL,RDB\n\t");
	// pull down WRB (LO act.)
	LCDCTRL &= ~(1 << WRB); // asm volatile ("cbr LCDCTRL,WRB\n\t");
	// enable outputs
	LCDDATADDR = 	(1<<D0) | (1<<D1) | (1<<D2) | 
					(1<<D3) | (1<<D4) | (1<<D5) | 
					(1<<D6) | (1<<D7);
}
// needed to activate and deativate the lcd write data bit //
inline void lcdDeactivateWRB (void)
{
	// never Deactivate WRB without switching off outputs!
	// set LCDDATADDR to input
	LCDDATADDR = 0;	// ~((1<<D0) | (1<<D1) | (1<<D2) | (1<<D3) | (1<<D4) | (1<<D5) | (1<<D6) | (1<<D7));
	// Pull up WRB (LO active signal)
	LCDCTRL |= (1 << WRB); // asm volatile ("sbr LCDCTRL,WRB\n\t");
}

// needed to activate and deativate the lcd read data bit /////////////////////
inline void lcdActivateRDB (void)
{
	// never activate RDB without switching off outputs!
	
	// Pull up write data bit (LO active signal)
	LCDCTRL |= (1 << WRB); // asm volatile ("sbr LCDCTRL,WRB\nt");
	
	// set LCDDATADDR to input
	LCDDATADDR = 	(0<<D0) | (0<<D1) | (0<<D2) | 
					(0<<D3) | (0<<D4) | (0<<D5) | 
					(0<<D6) | (0<<D7);

	// Pull down read data bit (LO active signal)
	LCDCTRL &= ~(1 << RDB); // asm volatile ("cbr LCDCTRL,RDB\n\t");

}

// needed to activate and deativate the lcd read data bit //
inline void lcdDeactivateRDB (void)
{
	// Pull up (LO active signal)
	LCDCTRL |= (1 << RDB); // asm volatile ("sbr LCDCTRL,RDB\n\t");
}

// needed to activate and deativate the lcd chip select bit ///////////////////
inline void lcdActivateCSB (void)
{
	// Pull down (LO active signal)
	LCDCTRL &= ~(1 << CSB); // asm volatile ("cbr LCDCTRL,CSB\n\t");
}
// needed to activate and deativate the lcd chip select bit //
inline void lcdDeactivateCSB (void)
{
	// Pull up (LO active signal)
	LCDCTRL |= (1 << CSB); // asm volatile ("sbr LCDDATA,CSB\n\t");
}


// needed to activate and deativate the lcd reset bit /////////////////////////
inline void lcdActivateReset (void)
{
	// Pull down (LO active signal)
	LCDCTRL &= ~(1 << RESB); // asm volatile ("cbr LCDCTLR,RESB\n\t");
}
// needed to activate and deativate the lcd reset bit //
inline void lcdDeactivateReset (void)
{
	// Pull up (LO active signal)
	LCDCTRL |= (1 << RESB); // asm volatile ("sbr LCDCTRL,RESB\n\t");
}

// needed to activate and deativate the lcd M86 bit ///////////////////////////
inline void lcdActivateM86 (void)
{
	// Pull up (HI active signal)
	LCDCTRL |= (1 << M86); // asm volatile ("sbr LCDCTRL,M86\n\t");
}
// needed to activate and deativate the lcd M86 bit //
inline void lcdDeactivateM86 (void)
{
	// Pull down (HI active signal)
	LCDCTRL &= ~(1 << M86); // asm volatile ("cbr LCDCTRL,M86\n\t");
}

// needed to activate and deativate the lcd RS bit ////////////////////////////
inline void lcdActivateRS (void)
{
	// Pull up (HI active signal)
	LCDCTRL |= (1 << RS); // asm volatile ("sbr LCDCTRL,RS\n\t");
}
// needed to activate and deativate the lcd RS bit //
inline void lcdDeactivateRS (void)
{
	// Pull down (HI active signal)
	LCDCTRL &= ~(1 << RS); // asm volatile ("cbr LCDCTRL,RS\n\t");
}
